(**************************************************************************)
(*  The CDuce interpreter                                                 *)
(*  Alain.Frisch@ens.fr                                                   *)
(*  Copyright 2003 (see LICENSE for details)                              *)
(**************************************************************************)

open Printf
open Pxp_document
open Schema_types

(* TODO when looking for xsd:{restriction,extension,...} has_element is used so
actually is possible that both more of them are provided.
IDEA: validate schema document using DTD for Schemas? *)

class type resolver =
  object
    method see : Schema_xml.schema_extension node -> unit
    method resolve_att : string -> att_decl
    method resolve_elt : now:bool -> string -> elt_decl ref
    method resolve_typ : now:bool -> string -> type_def ref
  end

exception Not_implemented of string

let debug = false
let debug_print s = if debug then prerr_endline s

let hashtbl_values tbl = Hashtbl.fold (fun _ valu acc -> valu :: acc) tbl []
let rec filter_out_none = function  (* not tail recursive *)
  | [] -> []
  | Some v :: tl -> v :: filter_out_none tl
  | None :: tl -> filter_out_none tl

let get_facet_nodes n = []  (* TODO facets *)
let restrict_simple_type base_type_def facets = base_type_def
let get_minOccurs n =
  try
    int_of_string n#extension#minOccurs
  with Not_found -> 1
let get_maxOccurs n =
  try
    (match n#extension#maxOccurs with
    | "unbounded" -> None
    | s -> Some (int_of_string s))
  with Not_found -> Some 1

let content_type_of_def = function
  | S def -> CT_simple def
  | C (CBuilt_in _) -> assert false
  | C (CUser_defined (_, _, _, _, _, ct)) -> ct

  (* check xsd:all constraints *)
let check_all_particle = function
  | 1, Some 1, All particles ->
      List.iter
        (function
          | _, Some 0, _ | _, Some 1, _ -> ()
          | _, _, _ ->
              raise (XSD_validation_error "all particles contained in a xsd:all
                term must have maxOccurs = {0, 1}"))
        particles
  | _, _, All _ ->
      raise (XSD_validation_error "xsd:all particles must have \
        minOccurs = maxOccurs = 1")
  | _, _, _ -> ()

let parse_facet (resolver: resolver) base_type_def n =
  debug_print "Schema_parser.parse_facet"; 
  let validate_base_type =
    Schema_validator.validate_simple_type base_type_def
  in
  let value =
    try
      n#extension#value
    with Not_found ->
      raise (XSD_validation_error "Missing required 'value' attribute")
  in
  let fixed =
    try
      bool_of_string n#extension#fixed
    with
    | Not_found -> false
    | Invalid_argument "bool_of_string" ->
        raise (XSD_validation_error (sprintf
          "Invalid value for 'fixed' attribute: '%s'" n#extension#fixed))
  in
  match n#node_type with
  | T_element "xsd:length" ->
      let length =
        Value.get_int (Schema_builtin.validate_nonNegativeInteger value)
      in
      F_length (length, fixed)
  | T_element "xsd:minLength" ->
      let length =
        Value.get_int (Schema_builtin.validate_nonNegativeInteger value)
      in
      F_minLength (length, fixed)
  | T_element "xsd:maxLength" ->
      let length =
        Value.get_int (Schema_builtin.validate_nonNegativeInteger value)
      in
      F_maxLength (length, fixed)
  | T_element "xsd:pattern" ->  (* TODO Schema regexp <> PCRE regexp :-(( *)
      F_pattern (Pcre.regexp value)
  | T_element "xsd:enumeration" ->
      F_enumeration (ValueSet.singleton (validate_base_type value))
  | T_element "xsd:whiteSpace" ->
      F_whiteSpace
        ((match value with
          | "collapse" -> WS_collapse
          | "preserve" -> WS_preserve
          | "replace" -> WS_replace
          | _ ->
              raise (XSD_validation_error (sprintf
                "'%s' isn't a valid whiteSpace value" value))),
        fixed)
  | T_element "xsd:maxInclusive" ->
      F_maxInclusive (validate_base_type value, fixed)
  | T_element "xsd:maxExclusive" ->
      F_maxExclusive (validate_base_type value, fixed)
  | T_element "xsd:minInclusive" ->
      F_minInclusive (validate_base_type value, fixed)
  | T_element "xsd:minExclusive" ->
      F_minExclusive (validate_base_type value, fixed)
  | T_element "xsd:totalDigits" ->
      let digits =
        Value.get_int (Schema_builtin.validate_positiveInteger value)
      in
      F_totalDigits (digits, fixed)
  | T_element "xsd:fractionDigits" ->
      let digits =
        Value.get_int (Schema_builtin.validate_nonNegativeInteger value)
      in
      F_fractionDigits (digits, fixed)
  | T_element unexpected ->
      raise (XSD_validation_error (sprintf "'%s' isn't a valid facet element"
        unexpected))
  | _ -> assert false

let parse_simple_type (resolver: resolver) n =
  debug_print "Schema_parser.parse_simple_type"; 
  if n#parent#node_type = T_element "xsd:schema" then
    resolver#see n;
  assert (n#node_type = T_element "xsd:simpleType");
  SBuilt_in "FAKE"  (* TODO facets *)
  (* FINQUI *)

  (** @return a value_constraint option from a attribute node *)
let constr_of_attr_node n validate =
  let (default, fixed) =
    ((try Some n#extension#default with Not_found -> None),
     (try Some n#extension#fixed with Not_found -> None))
 in
 try
   (match default, fixed with
   | Some v, None -> Some (Default (validate v))
   | None, Some v -> Some (Fixed (validate v))
   | None, None -> None
   | Some _, Some _ ->
       raise (XSD_validation_error ("Both 'default' and 'fixed' specified for \
        attribute " ^ n#extension#name)))
  with XSI_validation_error _ ->
    raise (XSD_validation_error ("Invalid value for constraint on \
      attribute " ^ n#extension#name))

let parse_att_decl (resolver: resolver) n =
  debug_print "Schema_parser.parse_att_decl"; 
  let name = n#extension#name in
  match n#parent#node_type with
  | T_element "xsd:schema" -> (* global attribute *)
      resolver#see n;
      let simple_type_def =
        (try
          parse_simple_type resolver (find_element "xsd:simpleType" n)
        with Not_found ->
          (try
            (match !(resolver#resolve_typ ~now:true n#extension#typ) with
            | S st -> st
            | C _ ->
                raise (XSD_validation_error
                  "Attributes can only assume simple type values"))
          with Not_found -> SBuilt_in "xsd:anySimpleType"))
      in
      let self_validate =
        Schema_validator.validate_simple_type simple_type_def
      in
      let constr = constr_of_attr_node n self_validate in
      name, simple_type_def, constr
  | _ -> assert false (* you have to use parse_attribute_use *)

  (** @return a attribute_use option. None means that the attribute is
  prohibited *)
let parse_attribute_use (resolver: resolver) n =
  debug_print "Schema_parser.parse_attribute_use"; 
  assert (n#node_type = T_element "xsd:attribute");
  let prohibited = try n#extension#prohibited with Not_found -> false in
  if prohibited then  (* attribute prohibited *)
    None
  else begin  (* attribute not prohibited *)
    let required = try n#extension#required with Not_found -> false in
    if n#extension#has_attribute "ref" then begin (* "ref" attribute *)
      let att_decl = resolver#resolve_att n#extension#ref in
      let simple_type_def = match att_decl with (_, std, _) -> std in
      let self_validate =
        Schema_validator.validate_simple_type simple_type_def
      in
      let constr = constr_of_attr_node n self_validate in
      Some (required, att_decl, constr)
    end else begin  (* no "ref" attribute *)
      let name = n#extension#name in
      let simple_type_def =
        (try
          parse_simple_type resolver (find_element "xsd:simpleType" n)
        with Not_found ->
          (try
            (match !(resolver#resolve_typ ~now:true n#extension#typ) with
            | S st -> st
            | C _ ->
                raise (XSD_validation_error
                  "Attributes can only assume simple type values"))
          with Not_found -> SBuilt_in "xsd:anySimpleType"))
      in
      let self_validate =
        Schema_validator.validate_simple_type simple_type_def
      in
      let att_decl = name, simple_type_def, None in
      let constr = constr_of_attr_node n self_validate in
      Some (required, att_decl, constr)
    end
  end

  (** @return a list of attribute uses from a xsd:restriction node wrt a base
  type definition *)
let attribute_uses_of_restriction ~(resolver: resolver) ~n ~base =
  (* BUG HERE *)
  let embedded =  (* associative list <name, attribute_use option> *)
    List.map
      (fun n ->
        let use = parse_attribute_use resolver n in
        let name =
          if n#extension#has_attribute "ref" then
            n#extension#ref
          else (* if n#extension#has_attribute "name" then *)
            n#extension#name
        in
        name, use)
      n#extension#find_attributes
  in
  (* /BUG HERE *)
  let from_base =
    match base with
    | C (CUser_defined (_, _, _, _, attribute_uses, _)) ->
        List.filter (* filters out attribute uses redefined and
                    prohibited in this type *)
          (fun use ->
            not (List.mem_assoc (name_of_attribute_use use) embedded))
          attribute_uses
    | _ -> []
  in
    (* remove prohibited from embedded attribute list *)
  filter_out_none (snd (List.split embedded)) @ from_base

  (** @return a list of attribute uses from a xsd:extension node wrt a base type
  definition *)
let attribute_uses_of_extension ~(resolver: resolver) ~n ~base =
  let embedded =  (* attribute_use option list *)
    List.map (parse_attribute_use resolver) n#extension#find_attributes
  in
  let from_base =
    match base with
    | C (CUser_defined (_, _, _, _, attribute_uses, _)) -> attribute_uses
    | _ -> []
  in
  filter_out_none embedded @ from_base

let counter = ref 0

let cuser_defined name base derivation attribute_uses ct =
  incr counter;
  CUser_defined (!counter, name, base, derivation, attribute_uses, ct)

let rec parse_complex_type (resolver: resolver) n =
  debug_print "Schema_parser.parse_complex_type"; 
  if n#parent#node_type = T_element "xsd:schema" then
    resolver#see n;
  let name = try Some n#extension#name with Not_found -> None in
  if n#extension#has_element "xsd:simpleContent" then begin
    let content = find_element "xsd:simpleContent" n in
    if content#extension#has_element "xsd:restriction" then begin
      (* simpleContent, restriction *)
      let restriction = find_element "xsd:restriction" content in
      let base = resolver#resolve_typ ~now:true restriction#extension#base in
      let attribute_uses =
        attribute_uses_of_restriction ~resolver ~n:restriction ~base:!base
      in
      let content_type =
        (match !base with
        | C (CUser_defined (_, _, _, _, _, (CT_simple base))) ->
            let base =
              try
                parse_simple_type resolver
                  (find_element "xsd:simpleType" restriction)
              with Not_found -> base
            in
            CT_simple (restrict_simple_type base (get_facet_nodes n))
        | _ -> assert false)
      in
      cuser_defined name !base Restriction attribute_uses content_type
    end else if content#extension#has_element "xsd:extension" then begin
      (* simpleContent, extension *)
      let extension = find_element "xsd:extension" content in
      let base = resolver#resolve_typ ~now:true extension#extension#base in
      let attribute_uses =
        attribute_uses_of_extension ~resolver ~n:extension ~base:!base
      in
      let content_type =
        (match !base with
        | C (CUser_defined (_, _, _, _, _, (CT_simple base))) -> CT_simple base
        | S simple_type_def -> CT_simple simple_type_def
        | _ -> assert false)
      in
      cuser_defined name !base Extension attribute_uses content_type
    end else
      (* simpleContent, neither extension nor restriction *)
      raise (XSD_validation_error "Neither <extension> nor <restriction> \
        element inside <simpleContent> element")
  end else if n#extension#has_element "xsd:complexContent" then begin
    let content = find_element "xsd:complexContent" n in
    if content#extension#has_element "xsd:restriction" then begin
      (* complexContent, restriction *)
      let restriction = find_element "xsd:restriction" content in
      let base = resolver#resolve_typ ~now:true restriction#extension#base in
      let attribute_uses =
        attribute_uses_of_restriction ~resolver ~n:restriction ~base:!base
      in
      let content_type =
        if restriction#extension#has_no_term then begin
          CT_empty
        end else begin
          let mixed =
            (try content#extension#mixed with Not_found ->
              (try restriction#extension#mixed with Not_found -> false))
          in
          (try
            let particle =
              parse_particle resolver restriction#extension#find_term
            in
            check_all_particle particle;
            CT_model (particle, mixed)
          with Not_found -> raise (XSD_validation_error "Can't find term"))
        end
      in
      cuser_defined name !base Restriction attribute_uses content_type
    end else if content#extension#has_element "xsd:extension" then begin
      (* complexContent, extension *)
      let extension = find_element "xsd:extension" content in
      let base = resolver#resolve_typ ~now:true extension#extension#base in
      let attribute_uses =
        attribute_uses_of_extension ~resolver ~n:extension ~base:!base
      in
      let content_type =
        let mixed =
          (try content#extension#mixed with Not_found ->
            (try extension#extension#mixed with Not_found -> false))
        in
        let base_ct = content_type_of_def !base in
        if extension#extension#has_no_term then
          base_ct
        else
          let term = lazy (
            try
              extension#extension#find_term
            with Not_found -> raise (XSD_validation_error ("Can't find term"))
          ) in
          match base_ct with
          | CT_empty ->
              let particle = parse_particle resolver (Lazy.force term) in
              check_all_particle particle;
              CT_model (particle, mixed)
          | CT_model (p, _) ->
              let base_particle = parse_particle resolver (Lazy.force term) in
              (match base_particle with
              | _, _, All _ ->
                  raise (XSD_validation_error ("Can't extend a content model, \
                    with an xsd:all particle"))
              | _ -> ());
              CT_model ((1, Some 1, Sequence (p :: [base_particle])), mixed)
          | CT_simple _ -> assert false
      in
      cuser_defined name !base Extension attribute_uses content_type
    end else
      (* complexContent, neither extension nor restriction *)
      raise (XSD_validation_error "Neither <extension> nor <restriction> \
        element inside <complexContent> element")
  end else begin
    (* neither simpleContent nor simpleContent, therefore ... *)
    (* ... complexContent, restriction: shortcut *)
    let base = resolver#resolve_typ ~now:true "xsd:anyType" in
    let attribute_uses =
      attribute_uses_of_restriction ~resolver ~n ~base:!base
    in
    let content_type =
      if n#extension#has_no_term then begin
        CT_empty
      end else begin
        let mixed = false in
        let particle = parse_particle resolver n#extension#find_term in
        check_all_particle particle;
        CT_model (particle, mixed)
      end
    in
    cuser_defined name !base Restriction attribute_uses content_type
  end

and parse_elt_decl (resolver: resolver) n =
  debug_print "Schema_parser.parse_elt_decl"; 
  match n#parent#node_type with
  | T_element "xsd:schema" -> (* global element *)
      resolver#see n;
      let name = n#extension#name in
      let type_def_ref =
        (try
          ref (S (parse_simple_type resolver (find_element "xsd:simpleType" n)))
        with Not_found ->
          (try
            ref (C (parse_complex_type resolver
              (find_element "xsd:complexType" n)))
          with Not_found ->
            resolver#resolve_typ ~now:false n#extension#typ))
      in
      name, type_def_ref, None
  | _ -> assert false (* you have to use parse_particle *)

and parse_particle (resolver: resolver) n =
  debug_print "Schema_parser.parse_particle"; 
  let (minOccurs, maxOccurs) = (get_minOccurs n, get_maxOccurs n) in
  match n#node_type with
  | T_element "xsd:element" when not (n#extension#has_attribute "ref") ->
      let name = n#extension#name in
      let type_def_ref =
        (try
          ref (S (parse_simple_type resolver (find_element "xsd:simpleType" n)))
        with Not_found ->
          (try
            ref (C (parse_complex_type resolver
              (find_element "xsd:complexType" n)))
          with Not_found ->
            resolver#resolve_typ ~now:false n#extension#typ))
      in
      minOccurs, maxOccurs, Elt (ref (name, type_def_ref, None))
  | T_element "xsd:element" (* when n#extension#has_attribute "ref" *)->
      let elt_decl = resolver#resolve_elt ~now:false n#extension#ref in
      minOccurs, maxOccurs, (Elt elt_decl)
  | T_element "xsd:all" ->
      minOccurs, maxOccurs,
      All (parse_particles resolver n#extension#find_terms)
  | T_element "xsd:sequence" ->
      minOccurs, maxOccurs,
      Sequence (parse_particles resolver n#extension#find_terms)
  | T_element "xsd:choice" ->
      minOccurs, maxOccurs,
      Choice (parse_particles resolver n#extension#find_terms)
  | _ -> assert false

  (* as above but used for "embedded" particle lists; it additionally check that
  the returned particle isn't an All one *)
and parse_particles (resolver: resolver) nodes =
  List.map
    (fun n ->
      match parse_particle resolver n with
      | _, _, All _ ->
          raise (XSD_validation_error "xsd:all particle can't appear here")
      | p -> p)
    nodes

module OrderedNode =
  struct
    type t = Schema_xml.schema_extension node
    let compare = Pxp_document.compare
  end
module NodeSet = Set.Make (OrderedNode)

  (* @param root schema document root node *)
class lazy_resolver =
  let fake_type_def = C (CBuilt_in "   FAKE TYP   ") in
  let fake_elt_decl = "   FAKE ELT   ", ref fake_type_def, None in
  let is_fake_type_def = (=) fake_type_def in
  let is_fake_elt_decl = (=) fake_elt_decl in
  let error_no_type_def name =
    raise (XSD_validation_error ("Can't find definition of type: " ^ name))
  in
  let error_no_elt_decl name =
    raise (XSD_validation_error ("Can't find declaration of element: " ^ name))
  in
  let error_no_att_decl name =
    raise (XSD_validation_error
      ("Can't find declaration of attribute: " ^ name))
  in
  fun root ->
  object (self)

    val typs = Hashtbl.create 17
    val attrs = Hashtbl.create 17
    val elts = Hashtbl.create 17

      (* register built-in types *)
    initializer
      List.iter (fun name -> Hashtbl.add typs name (ref (S (SBuilt_in name))))
        Schema_builtin.names

      (** seen nodes accounting *)

    val mutable seen_nodes = NodeSet.empty
    method already_seen n = NodeSet.mem n seen_nodes
    method see (n: Schema_xml.schema_extension node) =
      debug_print "lazy_resolver.see";
      if NodeSet.mem n seen_nodes then
        raise (XSD_validation_error "Types/Elements loop")
      else
        seen_nodes <- NodeSet.add n seen_nodes

      (** registration of global entities *)

    method register_typ name def =
      debug_print "lazy_resolver.register_typ";
      if (Hashtbl.mem typs name) &&
        (not (is_fake_type_def !(Hashtbl.find typs name)))
      then
        raise (XSD_validation_error ("Redefinition of type: " ^ name));
      if debug then
        (Format.fprintf Format.std_formatter
          "\nSchema_parser: registering TYPE %s:\n%a\n"
            name print_type def;
        Format.pp_print_flush Format.std_formatter ());
      let type_def_ref = self#resolve_typ ~now:false name in
      type_def_ref := def

    method register_elt name decl =
      debug_print "lazy_resolver.register_elt";
      if (Hashtbl.mem elts name) &&
        (not (is_fake_elt_decl !(Hashtbl.find elts name)))
      then
        raise (XSD_validation_error ("Redefinition of element: " ^ name));
      if debug then
        (Format.fprintf Format.std_formatter
          "\nSchema_parser: registering ELEMENT %s:\n%a\n"
            name print_elt_decl decl;
        Format.pp_print_flush Format.std_formatter ());
      let elt_decl_ref = self#resolve_elt ~now:false name in
      elt_decl_ref := decl

    method register_att name decl =
      debug_print "lazy_resolver.register_att";
      if Hashtbl.mem attrs name then
        raise (XSD_validation_error ("Redefinition of attribute: " ^ name));
      if debug then
        (Format.fprintf Format.std_formatter
          "\nSchema_parser: registering ATTRIBUTE %s:\n%a\n"
            name print_att_decl decl;
        Format.pp_print_flush Format.std_formatter ());
      Hashtbl.add attrs name decl

      (** entities lookup *)

    method resolve_typ ~now name =
      debug_print "lazy_resolver.resolve_typ";
      try
        Hashtbl.find typs name
      with Not_found ->
        if now then begin (* resolve now: look for global type definitions *)
          let node = root#extension#find_simpleType name in
          let typ_def =
            (try
              let node = node#extension#find_simpleType name in
              S (parse_simple_type (self :> resolver) node)
            with Not_found ->
              (try
                let node = node#extension#find_complexType name in
                C (parse_complex_type (self :> resolver) node)
              with Not_found -> error_no_type_def name))
          in
          Hashtbl.add typs name (ref typ_def)
        end else begin  (* resolve later: return a fake type ref *)
          Hashtbl.add typs name (ref fake_type_def)
        end;
        Hashtbl.find typs name

    method resolve_elt ~now name =
      debug_print "lazy_resolver.resolve_elt";
      try
        Hashtbl.find elts name
      with Not_found ->
        if now then begin (* resolve now: look for global element decls *)
          let node =
            try
              root#extension#find_global_element name
            with Not_found ->
              error_no_elt_decl name
          in
          let elt_decl = parse_elt_decl (self :> resolver) node in
          Hashtbl.add elts name (ref elt_decl)
        end else begin  (* resolve later: return fake element ref *)
          Hashtbl.add elts name (ref fake_elt_decl)
        end;
        Hashtbl.find elts name

    method resolve_att name =
      debug_print "lazy_resolver.resolve_att";
      try
        Hashtbl.find attrs name
      with Not_found ->
        let node =
          try
            root#extension#find_global_attribute name
          with Not_found -> error_no_att_decl name
        in
        let att_decl = parse_att_decl (self :> resolver) node in
        Hashtbl.add attrs name att_decl;
        att_decl

      (** acces to registered global entities *)

    method att_decls = hashtbl_values attrs
    method elt_decls =
      Hashtbl.fold
        (fun name decl acc -> (* check that all referenced elts are defined *)
          if is_fake_elt_decl !decl then
            error_no_elt_decl name
          else
            !decl :: acc)
        elts []
    method type_defs =
      Hashtbl.fold
        (fun name def acc -> (* check that all referenced types are defined *)
          if is_fake_type_def !def then
            error_no_type_def name
          else
            !def :: acc)
        typs []

  end

let parse_schema doc =
  let root = doc#root in
  let resolver = new lazy_resolver root in
  root#iter_nodes (fun n ->
    if not (resolver#already_seen n) then
      (match n#node_type with
      | T_element "xsd:element" ->
          resolver#register_elt n#extension#name
            (parse_elt_decl (resolver :> resolver) n)
      | T_element "xsd:simpleType" ->
          resolver#register_typ n#extension#name
            (S (parse_simple_type (resolver :> resolver) n))
      | T_element "xsd:complexType" ->
          resolver#register_typ n#extension#name
            (C (parse_complex_type (resolver :> resolver) n))
      | T_element "xsd:attribute" ->
          resolver#register_att n#extension#name
            (parse_att_decl (resolver :> resolver) n)
      | T_element e ->
          raise (XSD_validation_error ("Unexpected root element " ^ e))
      | _ -> ()));
  {
    type_defs = resolver#type_defs;
    att_decls = resolver#att_decls;
    elt_decls = resolver#elt_decls;
  }

