(**************************************************************************)
(*  The CDuce interpreter                                                 *)
(*  Alain.Frisch@ens.fr                                                   *)
(*  Copyright 2003 (see LICENSE for details)                              *)
(**************************************************************************)

open Printf

module StringMap = Map.Make (String)
module ValueSet = Set.Make (Value)
exception XSI_validation_error of string
exception XSD_validation_error of string

type derivation = Extension | Restriction

type value_constraint = Fixed of Value.t | Default of Value.t

type ws_handling = WS_preserve | WS_replace | WS_collapse

type facet =
  | F_length of int * bool
  | F_minLength of int * bool
  | F_maxLength of int * bool
  | F_pattern of Pcre.regexp
  | F_enumeration of ValueSet.t
  | F_whiteSpace of ws_handling * bool
  | F_maxInclusive of Value.t * bool
  | F_maxExclusive of Value.t * bool
  | F_minInclusive of Value.t * bool
  | F_minExclusive of Value.t * bool
  | F_totalDigits of int * bool
  | F_fractionDigits of int * bool

type simple_type_def =
  | SBuilt_in of string
  | SUser_defined of string option * variety * facet list * simple_type_def

and variety =
  | V_atomic of simple_type_def
  | V_list of simple_type_def
  | V_union of simple_type_def list

type att_decl = string * simple_type_def * value_constraint option
type attribute_use = bool * att_decl * value_constraint option

type term =
  | All of particle list
  | Choice of particle list
  | Sequence of particle list
  | Elt of elt_decl ref
and content_type =
  | CT_empty
  | CT_simple of simple_type_def
  | CT_model of particle * bool
  (* TODO move min and max directly on terms? *)
and particle = int * int option * term
and elt_decl = string * type_def ref * value_constraint option
and complex_type_def =
  | CBuilt_in of string
  | CUser_defined of
      int *
      string option * type_def * derivation *
      attribute_use list * content_type
and type_def = S of simple_type_def | C of complex_type_def

type schema = {
  type_defs: type_def list;
  att_decls: att_decl list;
  elt_decls: elt_decl list
}

let name_of_elt_decl (name, _, _) = name
let name_of_type_def = function
  | S (SBuilt_in name) -> name
  | C (CBuilt_in name) -> name
  | S (SUser_defined (Some name, _, _, _)) -> name
  | S (SUser_defined (None, _, _, _)) -> "| UNNAMED |"
  | C (CUser_defined (_, Some name, _, _, _, _)) -> name
  | C (CUser_defined (_, None, _, _, _, _)) -> "| UNNAMED |"
let name_of_attribute_use (_, (n, _, _), _) = n
let name_of_att_decl (n, _, _) = n

(* pretty printing *)

open Format

let print_simple_type ppf = function
  | SBuilt_in n -> fprintf ppf "%s" n
  | _ -> assert false
let print_att_decl ppf (name, st, constr) =
  fprintf ppf "@[{%s: %a}@]" name print_simple_type st
let rec print_elt_decl ppf (name, t, _) =
  fprintf ppf "@[%s%s:@ @[%a@]>@]" "<" name print_type !t
and print_type ppf = function
  | S s -> fprintf ppf "@[%a@]" print_simple_type s
  | C c -> fprintf ppf "@[%a@]" print_complex_type c
and print_complex_type ppf = function
  | CBuilt_in n -> fprintf ppf "@[%s@]" n
  | CUser_defined (_,_, _, _, _, ct) -> fprintf ppf "@[%a@]" print_ct ct
and print_ct ppf = function
  | CT_empty -> fprintf ppf "@[EMPTY@]"
  | CT_simple s -> print_simple_type ppf s
  | CT_model (p, _) -> print_particle ppf p
and print_particle ppf (min, max, term) =
  fprintf ppf "@[%a%s@]"
    print_term term
    (match min,max with
    | 1, Some 1 -> ""
    | 1, None -> "+"
    | 0, None -> "*"
    | min, None -> sprintf "{%d,*}" min
    | min, Some max -> sprintf "{%d,%d}" min max)
and print_term ppf = function
  | Elt elt_decl_ref -> print_elt_decl ppf !elt_decl_ref
  | Choice p ->
      fprintf ppf "@[[%t]@]"
        (fun ppf ->
          let rec aux = function
            | [] -> ()
            | hd::[] -> fprintf ppf "@[%a@]" print_particle hd
            | hd::tl -> fprintf ppf "@[%a@]@ | " print_particle hd; aux tl
          in
          aux p)
  | Sequence p ->
      fprintf ppf "@[[%t]@]"
        (fun ppf ->
          let rec aux = function
            | [] -> ()
            | hd::[] -> fprintf ppf "@[%a@]" print_particle hd
            | hd::tl -> fprintf ppf "@[%a@];@ " print_particle hd; aux tl
          in
          aux p)
  | _ -> assert false

let rec normalize_ws =
  let ws_RE = Pcre.regexp "[\t\r\n]" in
  let spaces_RE = Pcre.regexp "[ ]+" in
  fun handling s ->
  match handling with
  | WS_preserve -> s
  | WS_replace -> Pcre.replace ~rex:ws_RE ~templ:" " s
  | WS_collapse ->
      Pcre.replace ~rex:spaces_RE ~templ:" " (normalize_ws WS_replace s)

