
(** {2 Validation exceptions} *)

  (** validation failure in an _schema_ document *)
exception XSI_validation_error of string

  (** validation failure in an _instance_ document *)
exception XSD_validation_error of string

(** {2 Auxiliary modules} *)

  (* used to encode XML elements' attributes *)
module StringMap : Map.S with type key = string

  (* used to encode enumeration facet *)
module ValueSet : Set.S with type elt = Value.t

(** {2 XSD representation} *)

type derivation = Extension | Restriction

type value_constraint =
  | Fixed of Value.t
  | Default of Value.t

type ws_handling = WS_preserve | WS_replace | WS_collapse

  (* TODO int should be replaced by arbitrary long integers *)

type facet =
  | F_length of int * bool              (* length, fixed *)
  | F_minLength of int * bool           (* length, fixed *)
  | F_maxLength of int * bool           (* length, fixed *)
  | F_pattern of Pcre.regexp
  | F_enumeration of ValueSet.t
  | F_whiteSpace of ws_handling * bool  (* handling, fixed *)
  | F_maxInclusive of Value.t * bool    (* max, fixed *)
  | F_maxExclusive of Value.t * bool    (* max, fixed *)
  | F_minInclusive of Value.t * bool    (* min, fixed *)
  | F_minExclusive of Value.t * bool    (* min, fixed *)
  | F_totalDigits of int * bool         (* digits, fixed *)
  | F_fractionDigits of int * bool      (* digits, fixed *)

type simple_type_def =
  | SBuilt_in of string
  | SUser_defined of
      string option * (* name *)
      variety *
      facet list *
      simple_type_def (* base *)

and variety =
  | V_atomic of simple_type_def
  | V_list of simple_type_def
  | V_union of simple_type_def list

type att_decl =
  string *  (* name *)
  simple_type_def * (* type *)
  value_constraint option

type attribute_use =
  bool *  (* required *)
  att_decl *
  value_constraint option

type term =
  | All of particle list
  | Choice of particle list
  | Sequence of particle list
  | Elt of elt_decl ref

and content_type =
  | CT_empty
  | CT_simple of simple_type_def
  | CT_model of particle * bool (* true -> mixed; false -> element-only *)

and particle = int * int option * term  (* min, max (None = "*"), term *)

and elt_decl =
  string *  (* name *)
  type_def ref *  (* type *)
  value_constraint option

and complex_type_def =
  | CBuilt_in of string
  | CUser_defined of
      int *           (* unique id *)
      string option * (* name *)
      type_def *      (* base *)
      derivation *
      attribute_use list *
      content_type

and type_def =
  | S of simple_type_def
  | C of complex_type_def

type schema = {
  type_defs : type_def list;  (* global _type_ definitions *)
  att_decls : att_decl list;  (* global _attribute_ declarations *)
  elt_decls : elt_decl list;  (* global _element_ declarations *)
} 

(** {2 XSD printer *)

val print_elt_decl : Format.formatter -> elt_decl -> unit
val print_type : Format.formatter -> type_def -> unit
val print_simple_type : Format.formatter -> simple_type_def -> unit
val print_att_decl : Format.formatter -> att_decl -> unit
val print_complex_type : Format.formatter -> complex_type_def -> unit
val print_ct : Format.formatter -> content_type -> unit
val print_particle : Format.formatter -> particle -> unit
val print_term : Format.formatter -> term -> unit

(** {2 Misc} *)

val name_of_elt_decl : elt_decl -> string
val name_of_type_def : type_def -> string
val name_of_att_decl : att_decl -> string
val name_of_attribute_use : attribute_use -> string

  (** perform white space normalization according to a white space facet *)
val normalize_ws: ws_handling -> string -> string

