(**************************************************************************)
(*  The CDuce interpreter                                                 *)
(*  Alain.Frisch@ens.fr                                                   *)
(*  Copyright 2003 (see LICENSE for details)                              *)
(**************************************************************************)

let debug = false

open Printf
open Pxp_yacc
open Schema_types

exception Stop ;; (* internal *)

module OrderedStringOption =
  struct
    type t = string option
    let compare = Pervasives.compare
  end
  (* used to encode content model's "first". None value encode "epsilon" *)
module First = Set.Make (OrderedStringOption)

  (* a pair: validation function from stream to CDuce values, a first (usual
  FIRST construction on start tag names) *)
type validator = (Pxp_yacc.event Stream.t -> Value.t) * First.t
let fake_ct_validator: (((string * string) list -> Value.t) * validator) =
  ((fun _ -> assert false), ((fun _ -> assert false), First.empty))
let validate ~validator:(validate_fun, _) = validate_fun
let ct_validators = Hashtbl.create 17  (* complex type validators *)

module OrderedValidator =
  struct
    type t = bool * int * validator (* required, index, validator *)
    let compare (_, _, (_, f1)) (_, _, (_, f2)) = First.compare f1 f2
  end
  (* used to perform xsd:all validation *)
module ValidatorSet = Set.Make (OrderedValidator)

let empty_validator = (fun _ -> Value.sequence []), First.singleton (None)

  (* wrap a function validating a string with a validator *)
let pcdata_wrapper f = (fun stream -> f (Schema_xml.collect_pcdata stream))

let string_of_expect_token = function
  | `E_start_tag tag -> "<" ^ tag ^ ">"
  | `E_end_tag tag -> "</" ^ tag ^ ">"

let string_of_pxp_event = function
  | E_start_doc (version, standalone, dtd) -> "E_start_doc"
  | E_end_doc -> "E_end_doc"
  | E_start_tag (name, attlist, entity_id) -> sprintf "E_start_tag (%s)" name
  | E_end_tag (name, entity_id) -> sprintf "E_end_tag (%s)" name
  | E_char_data data ->
      sprintf "E_char_data (%s)" (Pcre.replace ~pat:"\n" ~templ:"\\n" data)
  | E_pinstr (target, value) -> "E_pinstr"
  | E_comment data -> "E_comment"
  | E_position (entity, line, col) -> "E_position"
  | E_error exn -> sprintf "E_error (%s)" (Pxp_types.string_of_exn exn)
  | E_end_of_stream -> "E_end_of_stream"
  | _ -> assert false

exception Not_a_start_tag of Pxp_yacc.event

  (* used, along with "first", to discriminate between choices *)
let rec peek_start_tag ?(ignore_ws = true) stream =
  match Schema_xml.peek stream with
  | E_char_data s when (Pxp_lib.only_whitespace s) && ignore_ws ->
      Stream.junk stream;
      peek_start_tag stream
  | E_pinstr (_, _) | E_comment _ | E_position _ | E_start_doc (_, _, _)
  | E_end_doc | E_error _ | E_end_of_stream ->
      Stream.junk stream;
      peek_start_tag stream
  | E_start_tag (name, _, _) -> name
  | e -> raise (Not_a_start_tag e)

  (* INVARIANT: if expect fails, the stream is left unchanged. Up to white space
  only CDATA removals if ignore_ws is true. *)
let expect ?(ignore_ws = true) expected stream =
  if debug then
    print_endline ("Expecting: " ^ string_of_expect_token expected);
  let rec junk () =
    match Schema_xml.peek stream with
    | E_char_data s when (Pxp_lib.only_whitespace s) && ignore_ws ->
        Stream.junk stream;
        junk ()
    | E_pinstr (_, _) | E_comment _ | E_position _ | E_start_doc (_, _, _)
    | E_end_doc | E_error _ | E_end_of_stream ->
        Stream.junk stream;
        junk ()
    | _ -> ()
  in
  junk ();
  match (expected, Schema_xml.peek stream) with
  | `E_start_tag tag, ((E_start_tag (t, _, _)) as found) when tag = t ->
      Stream.junk stream;
      found
  | `E_end_tag tag, ((E_end_tag (t, _)) as found) when tag = t ->
      Stream.junk stream;
      found
  | expected, found ->
      raise (XSI_validation_error (sprintf "expect failure: expected %s, found \
        %s" (string_of_expect_token expected) (string_of_pxp_event found)))

  (* TODO not tail recursive .... but Value.sequence isn't tail recursive too
  :-((( AND TODO terribly slow !!! *)
let flatten_cont =
  let rec dfv = function  (* depth first visit *)
    | Value.Pair (x, y) -> (dfv x) @ (dfv y)
    | v when Value.compare v Value.nil = 0 -> []
    | v -> [v]
  in
  fun l -> Value.sequence (dfv l)

let (>>=) n m =
  match (n, m) with Some n, m -> n >= m | None, _ -> true


let validator_of_simple_type = function
  | SBuilt_in s ->
      (pcdata_wrapper (Schema_builtin.__validate_fun_of_builtin s),
      First.empty)
  | SUser_defined (_, _, _, _) -> assert false

let validate_simple_type (simple_type_def: simple_type_def) value =
  validate ~validator:(validator_of_simple_type simple_type_def)
    ([< 'Pxp_yacc.E_char_data value; 'Pxp_yacc.E_end_of_stream >])

let validate_attrs_of_uses (attr_uses: attribute_use list) attrs =
  let attrs = (* (string * string) list -> (string StringMap.t) ref *)
    ref (List.fold_left (fun map (k,v) -> StringMap.add k v map)
      StringMap.empty attrs)
  in
  let record = Value.vrecord (List.fold_left
    (fun fields (required, (name, st, _), constr) ->
      try
        let value =
          try
            let value_raw = StringMap.find name !attrs in
            attrs := StringMap.remove name !attrs;
            let value = validate_simple_type st value_raw in
            (match constr with
            | None | Some (Default _) -> value
            | Some (Fixed v) when (Value.compare v value = 0)-> value
            | Some (Fixed _) ->
                raise (XSI_validation_error (sprintf "Value %s isn't \
                  compatible with 'fixed' constraint" value_raw)))
          with Not_found ->
            if required then
              raise (XSI_validation_error
                (sprintf "Required attribute '%s' is missing" name))
            else  (* optional *)
              (match constr with
              | None -> raise Stop
              | Some (Fixed v) | Some (Default v) -> v)
        in
        (name, value) :: fields
      with Stop -> fields)
    [] attr_uses)
  in
    (* remaining attributes in "attrs" are undeclared or prohibited *)
  match StringMap.fold (fun n _ acc -> n::acc) !attrs [] with
  | [] -> record
  | l ->
      raise (XSI_validation_error ("The following attributes are undeclared or \
        prohibited: " ^ String.concat ", " l))

let string_of_first ?(show_epsilon = false) first =
  let elts =
    First.fold
      (fun elt acc ->
        match elt with
        | None when show_epsilon -> " EPSILON " :: acc
        | None -> acc
        | Some e -> e :: acc)
      first []
  in
  String.concat ", " elts

let rec validator_of_particle (min, max, (term: term)) =
  assert ((min >= 0) && (max >>= min));
  let validator = validator_of_term term in
  let term_first = snd validator in
  let first =
    let old_first = snd validator in
    if min = 0 then First.add None old_first else old_first
  in
  match (min, max) with
  | (0, Some 0) -> empty_validator
  | (min, Some max) ->
      (fun stream ->
        let content = ref [] in
        for i = 1 to min do
          content := validate ~validator stream :: !content
        done;
        (try
          for i = 1 to max - min do
            let next = peek_start_tag stream in
            if not (First.mem (Some next) term_first) then
              raise Stop
            else
              content := validate ~validator stream :: !content
          done
        with Stop | Not_a_start_tag _ -> ());
        Value.sequence (List.rev !content)),
      first
  | (min, None) ->
      (fun stream ->
        let content = ref [] in
        for i = 1 to min do
          content := validate ~validator stream :: !content
        done;
        (try
          while true do
            let next = peek_start_tag stream in
            if not (First.mem (Some next) term_first) then
              raise Stop
            else
              content := validate ~validator stream :: !content
          done
        with Stop | Not_a_start_tag _ -> ());
        Value.sequence (List.rev !content)),
      first

and validator_of_term =
  let error first found =
    raise (XSI_validation_error (sprintf "Expected one of: %s; found %s"
      (string_of_first first) found))
  in
  function
  | All [] | Choice [] | Sequence [] -> empty_validator
  | All particles ->  (* assumption: maxOccurs = {0,1} *)
      let validators =
        snd (List.fold_left
          (fun (idx, acc) p ->
            match p with
            | 0, Some 1, _ | 0, Some 0, _ ->
                (idx + 1,
                ValidatorSet.add (false, idx, validator_of_particle p) acc)
            | 1, Some 1, _ ->
                (idx + 1,
                ValidatorSet.add (true, idx, validator_of_particle p) acc)
            | _ -> assert false)
          (0, ValidatorSet.empty) particles)
      in
      let len = ValidatorSet.cardinal validators in
      let first =
        ValidatorSet.fold (fun (_, _, (_, f)) acc -> First.union f acc)
          validators First.empty
      in
      (fun stream ->
        let values = Array.make len Value.Absent in
        let validators = ref validators in
        (try
          while true do
            let next = peek_start_tag stream in
            let (_, idx, validator) as elt = (* assumption: UPA *)
              ValidatorSet.choose (ValidatorSet.filter
                (fun (_, _, (_, f)) -> First.mem (Some next) f) !validators)
            in
            validators := ValidatorSet.remove elt !validators;
            values.(idx) <- validate ~validator stream
          done
        with
        | Not_found (* raised by "choose" *) | Not_a_start_tag _ -> ());
        let missing =
          ValidatorSet.filter (fun (req, _, _) -> req) !validators
        in
        if not (ValidatorSet.is_empty missing) then
          raise (XSI_validation_error ("incomplete xsd:all particle content, \
            expected (at least): " ^
            (string_of_first (match ValidatorSet.choose !validators with
            | _, _, (_, f) -> f))));
        Value.sequence (List.filter ((<>) Value.Absent)
          (Array.to_list values))),
      first

  | Choice particles ->
      (* TODO UPA *)
      let p_validators = List.map validator_of_particle particles in
      let find_validator name = (* find the validation function for a given
                                element *)
        let rec aux = function
          | [] -> raise Not_found
          | ((_, first) as v) :: tl when (First.mem (Some name) first) -> v
          | _ :: tl -> aux tl
        in
        aux p_validators
      in
      let first = (* union of choices' firsts *)
        List.fold_left (fun acc (_, f) -> First.union f acc) First.empty
          p_validators
      in
      (fun stream ->
        let next =
          try
            peek_start_tag stream
          with Not_a_start_tag ev ->
            error first (Schema_xml.string_of_pxp_event ev)
        in
        let validator =
          try find_validator next with Not_found -> error first next
        in
        validate ~validator stream),
      first
  | Sequence particles ->
      let p_validators = List.map validator_of_particle particles in
      let first = (* union of first until epsilon is in one of them *)
        let rec aux acc = function
          | [] -> acc
          | (_, first) :: tl ->
              let next_first = First.union acc first in
              if First.mem None first then aux next_first tl else next_first
        in
        aux First.empty p_validators
      in
      (fun stream ->
        let values = ref [] in
        List.iter
          (fun v -> values := validate ~validator:v stream :: !values)
          p_validators;
        Value.sequence (List.rev !values)),
      first
  | Elt decl -> validator_of_elt_decl !decl

and validator_of_complex_type' = function
  | CBuilt_in s ->  (* TODO uhm .... is this useful? *)
      ref
        ((fun _ -> assert false),
         (pcdata_wrapper (Schema_builtin.__validate_fun_of_builtin s),
          First.empty))
  | CUser_defined (id, _, _, _, attr_uses, ct) ->
      if not (Hashtbl.mem ct_validators id) then begin
        let reference = ref fake_ct_validator in
        Hashtbl.add ct_validators id reference;
        let validate_attrs = validate_attrs_of_uses attr_uses in
        let content_validator =
          match ct with
          | CT_empty -> (fun _ -> Value.sequence []), First.empty
          | CT_simple def -> validator_of_simple_type def
          | CT_model (particle, mixed) ->
              assert (not mixed); (* TODO mixed content support *)
              validator_of_particle particle
        in
        reference := (validate_attrs, content_validator)
      end;
      Hashtbl.find ct_validators id

and validator_of_elt_decl ((name, def, _): elt_decl) = (* TODO constraints *)
  let first = First.singleton (Some name) in
  match !def with
  | S def ->
      let validator = validator_of_simple_type def in
      (fun stream ->
        ignore (expect (`E_start_tag name) stream);
        let cont = validate ~validator stream in
        ignore (expect (`E_end_tag name) stream);
        Value.Xml (Value.Atom (Atoms.mk_ascii name), Value.vrecord [], cont)),
      first
  | C def ->
      let reference = validator_of_complex_type' def in
      (fun stream ->
        match expect (`E_start_tag name) stream with
        | E_start_tag (name, attrs, _) ->
            let (validate_attrs, validator) = !reference in
            let attrs = validate_attrs attrs in
            let cont = flatten_cont (validate ~validator stream) in
            ignore (expect (`E_end_tag name) stream);
            Value.Xml (Value.Atom (Atoms.mk_ascii name), attrs, cont)
        | _ -> assert false),
      first

