(**************************************************************************)
(*  The CDuce interpreter                                                 *)
(*  Alain.Frisch@ens.fr                                                   *)
(*  Copyright 2003 (see LICENSE for details)                              *)
(**************************************************************************)

open Pxp_document
open Pxp_yacc
open Pxp_types

class schema_extension =
  let is_ncname = (* TODO check if s is a NCName, this is only a hack that
                  checks if no spaces are presents *)
    let space_RE = Pcre.regexp " " in
    fun s -> not (Pcre.pmatch ~rex:space_RE s)
  in
  object (self)

    val mutable node = (None: schema_extension node option)

    method clone = {< >} 
    method node = match node with None -> assert false | Some n -> n
    method set_node n = node <- Some n

    method private string_attribute name =  (* TODO check if we should accept
                                            also Valuelist attributes *)
      match self#node#attribute name with Value v -> v | _ -> raise Not_found
    method private ncattr name = self#string_attribute name

    method base = self#ncattr "base"
    method default = self#ncattr "default"
    method fixed = self#ncattr "fixed"
    method maxOccurs = self#ncattr "maxOccurs"
    method minOccurs = self#ncattr "minOccurs"
    method mixed = bool_of_string (self#ncattr "mixed")
    method name = self#ncattr "name"
    method ref = self#ncattr "ref"
    method typ = self#ncattr "type"
    method value = self#ncattr "value"

    method prohibited =
      try
        match self#ncattr "use" with
        | "prohibited" -> true
        | _ -> false
      with Not_found -> false

    method required =
      try
        match self#ncattr "use" with
        | "required" -> true
        | _ -> false
      with Not_found -> false

    method has_attribute a = List.mem a (self#node#attribute_names)
    method has_element e =
      try
        ignore (find_element e self#node); true
      with Not_found -> false

    method find_facets =
      let facets = ref [] in
      self#node#iter_nodes (fun n -> match n#node_type with
        | T_element "length" | T_element "minLength" | T_element "maxLength"
        | T_element "pattern" | T_element "enumeration" | T_element "whiteSpace"
        | T_element "maxInclusive" | T_element "maxExclusive"
        | T_element "minInclusive" | T_element "minExclusive"
        | T_element "totalDigits" | T_element "fractionDigits" ->
            facets := n :: !facets
        | _ -> ());
      !facets

    method find_terms =
      List.filter
        (fun n ->
          match n#node_type with
          | T_element "xsd:all"
          | T_element "xsd:choice"
          | T_element "xsd:sequence"
          | T_element "xsd:element" -> true
          | _ -> false)
        self#node#sub_nodes

    method find_term =
      match self#find_terms with
      | [t] -> t
      | _ -> raise Not_found

    method find_attributes =
      List.filter
        (fun n ->
          match n#node_type with
          | T_element "xsd:attribute" -> true
          | _ -> false)
        self#node#sub_nodes

    method private find_gen tagname name =
      match
        find_all
          (fun n ->
            (n#node_type = T_element tagname) &&
            (try n#extension#name = name with Not_found -> false))
          self#node
      with
      | [t] -> t
      | [] -> raise Not_found
      | _ -> assert false

    method find_simpleType = self#find_gen "xsd:simpleType"
    method find_complexType = self#find_gen "xsd:complexType"
    method find_global_element = self#find_gen "xsd:element"
    method find_global_attribute = self#find_gen "xsd:attribute"

    method has_no_term = self#find_terms = []

    method sub_elements =
      List.filter
        (fun n -> match n#node_type with T_element _ -> true | _ -> false)
        self#node#sub_nodes
    method nth_element = List.nth self#sub_elements

  end

let spec =
  make_spec_from_alist
    ~data_exemplar:             (new data_impl (new schema_extension))
    ~default_element_exemplar:  (new element_impl (new schema_extension))
    ~element_alist:             []
    ()

let pxp_tree_of fname =
  parse_wfdocument_entity default_config (from_file fname) spec

let pxp_stream_of_file fname =
  let config = { default_config with drop_ignorable_whitespace = true } in
  let entity_manager =
    create_entity_manager ~is_document:true config (from_file fname)
  in
  let pull_parser =
    create_pull_parser default_config
      (`Entry_document [ `Extend_dtd_fully; `Parse_xml_decl ])
      entity_manager
  in
  Stream.from pull_parser

class foo_entity_id = object end
let eid = new foo_entity_id
type to_be_visited =
  | Fully of Value.t  (* xml values still to be visited *)
  | Half of Value.t   (* xml values half visited (i.e. E_start_tag generated) *)
  | Other of Value.t  (* other values *)

let pxp_stream_of_value v =
  let stack = ref [Fully v] in
  let f _ = (* lazy visit of a tree of CDuce XML values, stack keeps track of
            what is still to be visited *)
    (match !stack with
    | (Fully ((Value.Xml (Value.Atom a, attrs, _)) as v)) :: tl ->
	let (ns,a) = Atoms.value a in
	assert( ns == Ns.empty );
        let tag_ascii = Encodings.Utf8.to_string a in
        let attrs_ascii =
          List.map (fun (k, v) -> (k, Value.get_string_latin1 v))
            (Value.get_fields attrs)
        in
        let event = Some (E_start_tag (tag_ascii, attrs_ascii, eid)) in
        stack := (Half v) :: tl;
        let children = ref [] in  (* TODO inefficient *)
        let push v = children := v :: !children in
        Value.iter_xml
          (fun pcdata -> push (Other (Value.string_utf8 pcdata)))
          (fun v ->
            match v with
            | (Value.Xml (_, _, _)) as v -> push (Fully v)
            | v -> raise (Invalid_argument "Schema_xml.pxp_stream_of_value"))
          v;
        stack := (List.rev !children) @ !stack;
        event
    | (Half (Value.Xml (Value.Atom a, _, _))) :: tl ->
	let (ns,a) = Atoms.value a in
	assert( ns == Ns.empty );
        let tag_ascii = Encodings.Utf8.to_string a in
        let event = Some (E_end_tag (tag_ascii, eid)) in
        stack := tl;
        event
    | (Fully (Value.Xml (_, _, _)))::_ | (Half (Value.Xml (_, _, _)))::_ ->
        failwith "Schema_xml.pxp_stream_of_value: non-atom-tag xml value"
    | (Other v) :: tl ->
        let event = Some (E_char_data (Value.get_string_latin1 v)) in
        stack := tl;
        event
    | [] -> None
    | _ -> assert false)
  in
  Stream.from f

open Printf

let string_of_pxp_event = function
  | E_start_doc (version, standalone, dtd) -> "E_start_doc"
  | E_end_doc -> "E_end_doc"
  | E_start_tag (name, attlist, entity_id) -> sprintf "E_start_tag (%s)" name
  | E_end_tag (name, entity_id) -> sprintf "E_end_tag (%s)" name
  | E_char_data data ->
      sprintf "E_char_data (%s)" (Pcre.replace ~pat:"\n" ~templ:"\\n" data)
  | E_pinstr (target, value) -> "E_pinstr"
  | E_comment data -> "E_comment"
  | E_position (entity, line, col) -> "E_position"
  | E_error exn -> sprintf "E_error (%s)" (Pxp_types.string_of_exn exn)
  | E_end_of_stream -> "E_end_of_stream"
  | _ -> assert false

let rec dump_stream s =
  print_endline (string_of_pxp_event (Stream.next s));
  flush stdout;
  dump_stream s

let dump_stream s = try dump_stream s with Stream.Failure -> ()

  (* peek version that assume the stream isn't at the end *)
let peek s =
  match Stream.peek s with
  | Some v -> v
  | None -> raise Stream.Failure

  (* collect all E_char_data events from a PXP stream and return the
  concatenation of their datas *)
let collect_pcdata s =
  let buf = Buffer.create 1 in
  let rec collect () =
    match peek s with
    | E_char_data d ->
        Buffer.add_string buf d; Stream.junk s; collect ()
    | _ -> Buffer.contents buf
  in
  collect ()

