(**************************************************************************)
(*  The CDuce interpreter                                                 *)
(*  Alain.Frisch@ens.fr                                                   *)
(*  Copyright 2003 (see LICENSE for details)                              *)
(**************************************************************************)

(* Optimization ideas:
    -    (A|B) & (C|D) = A | (B & (C|D))    if A <= C
*)

module type S =
sig
  type 'a elem
  type 'a t

  val equal : 'a t -> 'a t -> bool
  val compare: 'a t -> 'a t -> int
  val hash: 'a t -> int

  external get: 'a t -> ('a elem list * 'a elem list) list = "%identity"

  val empty : 'a t
  val full  : 'a t
  val cup   : 'a t -> 'a t -> 'a t
  val cap   : 'a t -> 'a t -> 'a t
  val diff  : 'a t -> 'a t -> 'a t
  val atom  : 'a elem -> 'a t

  val map : ('a elem-> 'b elem) -> 'a t -> 'b t
  val iter: ('a elem -> unit) -> 'a t -> unit
  val compute: empty:'d -> full:'c -> cup:('d -> 'c -> 'd) 
    -> cap:('c -> 'b -> 'c) -> diff:('c -> 'b -> 'c) ->
    atom:('a elem -> 'b) -> 'a t -> 'd
  val compute_bool: ('a elem -> 'b t) -> 'a t -> 'b t
    
  val print: string -> (Format.formatter -> 'a elem -> unit) -> 'a t ->
    (Format.formatter -> unit) list
    

  val trivially_disjoint : 'a t -> 'a t -> bool    
  val check: 'a t -> unit
end

module Make(X : SortedList.ARG) = struct
  type 'a elem = 'a X.t
  module SList = SortedList.Make_transp(X)
  module SSList = SortedList.Make_transp
    (struct
       type 'a t = 'a SList.t * 'a SList.t
       let compare (x1,y1) (x2,y2) =
	 let c = SList.compare x1 x2 in if c <> 0 then c
	 else SList.compare y1 y2
       let equal (x1,y1) (x2,y2) =
	 (SList.equal x1 x2) && (SList.equal y1 y2)
       let hash (x,y) =
	 SList.hash x + 17 * SList.hash y
     end)
  type 'a t = 'a SSList.t
  let hash = SSList.hash
  let compare = SSList.compare
  let equal = SSList.equal

  external get: 'a t -> ('a elem list * 'a elem list) list = "%identity"


let empty = [ ]
let full  = [ [],[] ]

let atom x = [ [x],[] ]

let may_remove (p1,n1) (p2,n2) =
(*  false *)
  (SList.subset p2 p1) && (SList.subset n2 n1)

(* in some cases, it is faster to avoid may_remove...
   investigate this... *)


let cup t s = 
  if t == s then t
  else match (t,s) with
    | [],s -> s
    | t,[] -> t
    | [ [], [] ], _ | _, [ [], [] ] -> full
    | _ ->
	let s=
	  SSList.filter 
	    (fun (p,n) -> not (SSList.exists (may_remove (p,n)) t)) s in
	let t=
	  SSList.filter 
	    (fun (p,n) -> not (SSList.exists (may_remove (p,n)) s)) t in 
	SSList.cup s t

(*
let clean accu t =
  let rec aux accu = function
    | (p,n) :: rem ->
	if (List.exists (may_remove (p,n)) accu)
	  || (List.exists (may_remove (p,n)) rem)
	then aux accu rem
	else aux ((p,n)::accu) rem 
    | [] -> accu
  in
  SSList.from_list (aux accu t)
*)

    

let rec fold2_aux f a x = function
  | [] -> x
  | h :: t -> fold2_aux f a (f x a h) t

let rec fold2 f x l1 l2 =
  match l1 with
    | [] -> x
    | h :: t -> fold2 f (fold2_aux f h x l2) t l2

let rec should_add x = function
  | [] -> true
  | y::rem -> if may_remove x y then false else should_add x rem

let rec clean_add accu x = function
  | [] -> accu
  | y::rem -> 
      if may_remove y x then clean_add accu x rem 
      else clean_add (y::accu) x rem

let cap s t =
  if s == t then s
  else if s == full then t
  else if t == full then s
  else if (s == empty) || (t == empty) then empty
  else
    let (lines1,common,lines2) = SSList.split s t in
    let rec aux lines (p1,n1) (p2,n2) =
        if (SList.disjoint p1 n2) && (SList.disjoint p2 n1)
        then 
	  let x = (SList.cup p1 p2, SList.cup n1 n2) in
	  if should_add x lines then clean_add [x] x lines else lines
	else lines
    in
    SSList.from_list 
      (fold2 aux (SSList.get common) (SSList.get lines1) (SSList.get lines2))

let diff c1 c2 =
  if (c2 == full) || (c1 == c2) then empty
  else if (c1 == empty) || (c2 == empty) then c1
  else
    let c1 = SSList.diff c1 c2 in
    let line (p,n) =
      let acc = SList.fold (fun acc a -> ([], [a]) :: acc) [] p in
      let acc = SList.fold (fun acc a -> ([a], []) :: acc) acc n in
      SSList.from_list acc
    in
    SSList.fold (fun c1 l -> cap c1 (line l)) c1 c2


let rec map f t =
  let lines =
    List.fold_left
      (fun lines (p,n) ->
	 let p = SList.map f p and n = SList.map f n in
	 if (SList.disjoint p n) then (p,n) :: lines else lines)
      []
      t
  in
  SSList.from_list lines

let iter f t =
  SSList.iter (fun (p,n) -> SList.iter f p; SList.iter f n) t

let compute ~empty ~full ~cup ~cap ~diff ~atom t =
  let line (p,n) =
    List.fold_left (fun accu x -> diff accu (atom x)) (
      List.fold_left (fun accu x -> cap accu (atom x)) full p
    ) n in
  List.fold_left (fun accu l -> cup accu (line l)) empty t
  
let compute_bool f = 
  compute ~empty ~full ~cup ~cap ~diff ~atom:f


let print any f =
  List.map 
    (function 
	 (p1::p,n) -> 
	   (fun ppf ->
	      Format.fprintf ppf "@[%a" f p1;
	      List.iter (fun x -> Format.fprintf ppf " &@ %a" f x) p;
	      List.iter (fun x -> Format.fprintf ppf " \\@ %a" f x) n;
	      Format.fprintf ppf "@]";
	   )
       | ([],[]) -> 
	   (fun ppf -> Format.fprintf ppf "%s" any)
       | ([],[n]) ->
	   (fun ppf -> Format.fprintf ppf "@[%s \\ %a@]" any f n)
       | ([],n1::n) ->
	   (fun ppf -> 
              Format.fprintf ppf "@[%s" any;
	      List.iter (fun x -> Format.fprintf ppf " \\@ %a" f x) n;
	      Format.fprintf ppf "@]";
	   )
    )

let trivially_disjoint a b = cap a b = []

let check b = ()
(*
  SSList.check b;
  SSList.iter 
    (fun (p,n) -> 
       SList.check p; 
       SList.check n; 
       assert (SList.disjoint p n)
    )
    b
*)

end

include Make(
  struct 
    type 'a t = 'a 
    let hash = Hashtbl.hash 
    let equal x y = x = y 
    let compare = compare 
  end)
