(**************************************************************************)
(*  The CDuce interpreter                                                 *)
(*  Alain.Frisch@ens.fr                                                   *)
(*  Copyright 2003 (see LICENSE for details)                              *)
(**************************************************************************)

(* Typed abstract syntax *) 

(*  Some sub-expression may have to be type-checked several times.
    We first build the ``skeleton'' of the typed ast
    (basically the parsed ast with types and patterns replaced with there
    internal representation), then type check it.

    The exp_typ and br_typ fields are updated to capture all the possible
    values than can result from the expression or flow to the branch
*)

open Location
open Ident

type tpat = Patterns.node
type ttyp = Types.node

type texpr  = 
    { exp_loc : loc; 
      mutable exp_typ : Types.descr; 
      exp_descr : texpr';
    }
and  texpr' = 
  | Forget of texpr * ttyp
  (* CDuce is a Lambda-calculus ... *)
  | Var of id
  | Apply of texpr * texpr
  | Abstraction of abstr
      
  (* Data constructors *)
  | Cst of Types.const
  | Pair of texpr * texpr
  | Xml of texpr * texpr
  | RecordLitt of texpr label_map
  | String of U.uindex * U.uindex * U.t * texpr
      
  (* Data destructors *)
  | Match of texpr * branches
  | Map of texpr * branches
  | Transform of texpr * branches
  | Xtrans of texpr * branches
  | Validate of texpr * string * string (* expression, schema, schema type *)
  | RemoveField of texpr * label
  | Dot of texpr * label
      
  (* Exception *)
  | Try of texpr * branches

  | UnaryOp of unary_op * texpr
  | BinaryOp of binary_op * texpr * texpr

and unary_op = {
  un_op_typer : loc -> typ_fun -> typ_fun;
  un_op_eval  : Value.t -> Value.t
}
and binary_op = {
  bin_op_typer : loc -> typ_fun -> typ_fun -> typ_fun;
  bin_op_eval  : Value.t -> Value.t -> Value.t
}
and typ_fun = Types.descr -> bool -> Types.descr

and abstr = { 
  fun_name : id option; 
  fun_iface : (Types.descr * Types.descr) list;
  fun_body : branches;
  fun_typ  : Types.descr;
  fun_fv   : fv
}

and let_decl = {
  let_pat : tpat;
  let_body : texpr;
  mutable let_compiled : 
    (Patterns.Compile.dispatcher * int id_map) option
}

and branches = { 
  mutable br_typ : Types.descr; (* Type of values that can flow to branches *)
  br_accept : Types.descr;  (* Type accepted by all branches *)
  br_branches: branch list;

  mutable br_compiled : compiled_branches option;
}
and branch = { 
  br_loc : loc;
  mutable br_used : bool; 
  br_pat : tpat; 
  br_body :  texpr 
}
and compiled_branches =  
    Patterns.Compile.dispatcher * texpr Patterns.Compile.rhs array


let dispatcher brs =
  match brs.br_compiled with
    | Some d -> d
    | None ->
	let aux b = b.br_pat, b.br_body in
	let x = Patterns.Compile.make_branches
		  brs.br_typ
		  (List.map aux brs.br_branches) in
	brs.br_compiled <- Some x;
	x
		     
let dispatcher_let_decl l =
  match l.let_compiled with
    | Some d -> d
    | None ->
	let comp = Patterns.Compile.make_branches
		     (Types.descr (Patterns.accept l.let_pat))
		     [ l.let_pat, () ]  in
	let x = match comp with
	  | (disp, [| Patterns.Compile.Match (l, ()) |]) -> (disp,l)
	  | _ -> assert false
	in
	l.let_compiled <- Some x;
	x

